"""Module psiturk_config."""
from __future__ import generator_stop
import os
from configparser import ConfigParser
from dotenv import load_dotenv, find_dotenv
from .psiturk_exceptions import EphemeralContainerDBError, PsiturkException

class PsiturkConfig(ConfigParser):
    """PsiturkConfig class."""

    def __init__(self, local_config="config.txt",
                 global_config_name=".psiturkconfig", **kwargs):
        """Init."""
        load_dotenv(find_dotenv(usecwd=True))
        if 'PSITURK_GLOBAL_CONFIG_LOCATION' in os.environ:
            global_config = os.path.join(
                os.environ['PSITURK_GLOBAL_CONFIG_LOCATION'], global_config_name)
        else:  # if nothing is set default to user's home directory
            global_config = "~/" + global_config_name
        self.parent = ConfigParser
        super().__init__(**kwargs)
        self.local_file = local_config
        self.global_file = os.path.expanduser(global_config)

    def load_config(self):
        """Load config."""
        defaults_folder = os.path.join(
            os.path.dirname(__file__), "default_configs")
        local_defaults_file = os.path.join(
            defaults_folder, "local_config_defaults.txt")
        global_defaults_file = os.path.join(
            defaults_folder, "global_config_defaults.txt")
        cloud_defaults_file = os.path.join(
            defaults_folder, "cloud_config_defaults.txt")

        # Read files in this order, with later settings overriding
        # earlier ones:
        #
        # * global default
        # * local default
        # * if ON_CLOUD, cloud defaults
        # * user's global file
        # * user's local's file
        # * env vars
        self.read_file(open(global_defaults_file))
        self.read_file(open(local_defaults_file))

        # Backwards compatibility
        if 'ON_HEROKU' in os.environ:
            os.environ['ON_CLOUD'] = '1'

        # read in default cloud config
        if 'ON_CLOUD' in os.environ:
            self.read_file(open(cloud_defaults_file))

        self.read([self.global_file, self.local_file])

        # backwards compatibility
        backwards_compatibilities = [
            {
                'in_section': 'Server Parameters',
                'prefer_this': 'errorlog',
                'over_this': 'logfile'
            }
        ]
        for bc in backwards_compatibilities:
            if (not self.has_option(bc['in_section'], bc['prefer_this'])) and self.has_option(bc['in_section'], bc['over_this']):
                self.set(bc['in_section'], bc['prefer_this'], self.get(bc['in_section'], bc['over_this']))
            env_key = f'PSITURK_{bc["over_this"].upper()}'
            if env_key in os.environ:
                self.set(bc['in_section'], bc['prefer_this'], os.environ.get(env_key))

        # prefer environment
        these_as_they_are = ['PORT', 'DATABASE_URL']  # heroku sets these
        for section in self.sections():
            for config_var in self[section]:
                config_var_upper = config_var.upper()
                config_val_env_override = None

                if config_var_upper in these_as_they_are:
                    config_val_env_override = os.environ.get(config_var_upper)

                # prefer any `PSITURK_` key even over `these_as_they_are` variants
                psiturk_key = f'PSITURK_{config_var_upper}'
                if psiturk_key in os.environ:
                    config_val_env_override = os.environ.get(psiturk_key)

                if config_val_env_override:
                    self.set(section, config_var, config_val_env_override)

        # Also set bc's in reverse direction, for code compatibility
        for bc in backwards_compatibilities:
            self.set(bc['in_section'], bc['over_this'], self.get(bc['in_section'], bc['prefer_this']))

        # heroku files are ephemeral.
        # Error if we're trying to use a file as the db
        if 'ON_CLOUD' in os.environ:
            database_url = self.get('Database Parameters',
                                    'database_url')
            if ('localhost' in database_url) or ('sqlite' in database_url):
                raise EphemeralContainerDBError(database_url)

    def get_ad_url(self):
        """Get ad url."""
        if self.has_option('HIT Configuration', 'ad_url'):
            return self.get('HIT Configuration', 'ad_url')
        else:
            need_these = ['ad_url_domain', 'ad_url_protocol', 'ad_url_port',
                          'ad_url_route']
            for need_this in need_these:
                if not self.get('HIT Configuration', need_this):
                    raise PsiturkException(
                        message=f'missing ad_url config var `{need_this}`')
            ad_url_domain = self.get('HIT Configuration', 'ad_url_domain')
            ad_url_protocol = self.get('HIT Configuration', 'ad_url_protocol')
            ad_url_port = self.get('HIT Configuration', 'ad_url_port')
            ad_url_route = self.get('HIT Configuration', 'ad_url_route')
            return f"{ad_url_protocol}://{ad_url_domain}:{ad_url_port}/{ad_url_route}"



from __future__ import generator_stop

class PsiturkException(Exception):
    def __init__(self, *args, **kwargs):
        super(PsiturkException, self).__init__(*args)
        self.message = kwargs['message'] if 'message' in kwargs else ''
        
    def __str__(self):
        return '{}: {}'.format(type(self).__name__, self.message)
        
    def to_dict(self):
        return {
            'exception': type(self).__name__,
            'message': self.message
        }

####################################################
# AMT Services Exceptions
####################################################


class AmtServicesException(PsiturkException):
    
    def __init__(self, **kwargs):
        super(AmtServicesException, self).__init__(**kwargs)


class AWSAccessKeysNotSetError(AmtServicesException):
    def __init__(self, **kwargs):
        message = 'AWS access keys not set in ~/.psiturkconfig; please enter valid credentials.'
        super(AWSAccessKeysNotSetError, self).__init__(message=message, **kwargs)


class NoMturkConnectionError(AmtServicesException):
    def __init__(self, **kwargs):
        message = 'Sorry, unable to connect to Amazon Mechanical Turk. AWS credentials invalid.'
        super(NoMturkConnectionError, self).__init__(message=message, **kwargs)

####################################################
# AMT Service Wrapper Exceptions
####################################################


class AmtServicesWrapperError(PsiturkException):
    pass


class DoBonusError(AmtServicesWrapperError):
    def __init__(self, assignment_id='', **kwargs):
        super(DoBonusError, self).__init__(**kwargs)
        self.assignment_id = assignment_id


class AssignmentAlreadyBonusedError(DoBonusError):
    pass


class BadBonusAmountError(DoBonusError):
    def __init__(self, amount, **kwargs):
        super(BadBonusAmountError, self).__init__(**kwargs)
        self.message = 'Bonus must be a number greater than 0. You gave: {}'.format(
            amount)


class BonusReasonMissingError(DoBonusError):
    pass


class NoAutoBonusAmountSetError(DoBonusError):
    pass


class AssignmentIdNotFoundInLocalDBError(AmtServicesWrapperError):
    def __init__(self, assignment_id='', **kwargs):
        super(AssignmentIdNotFoundInLocalDBError, self).__init__(**kwargs)
        self.assignment_id = assignment_id


class WorkerIdNotFoundInLocalDBError(AmtServicesWrapperError):
    pass


class MissingArgumentsError(AmtServicesWrapperError):
    pass


class InvalidPsiturkCredentialsError(AmtServicesWrapperError):
    def __init__(self, *args, **kwargs):
        super(InvalidPsiturkCredentialsError, self).__init__(*args, **kwargs)
        self.message = '\n'.join([
            '*****************************',
            '  Sorry, your psiTurk Credentials are invalid.\n ',
            '  You cannot create ads and hits until you enter valid credentials in ',
            '  the \'psiTurk Access\' section of ~/.psiturkconfig.  You can obtain your',
            '  credentials or sign up at https://www.psiturk.org/login.\n'])


class InvalidAWSCredentialsError(AmtServicesWrapperError):
    def __init__(self, *args, **kwargs):
        super(InvalidAWSCredentialsError, self).__init__(*args, **kwargs)
        self.message = '\n'.join([
            '*****************************',
            '  Sorry, your AWS Credentials are invalid.\n ',
            '  You cannot create ads and hits until you enter valid credentials in ',
            '  the \'AWS Access\' section of ~/.psiturkconfig.  You can obtain your ',
            '  credentials via the Amazon AMT requester website.\n'])


class AdPsiturkOrgError(AmtServicesWrapperError):
    pass


class AdHtmlNotFoundError(AmtServicesWrapperError):
    def __init__(self, *args, **kwargs):
        super(AdHtmlNotFoundError, self).__init__(*args, **kwargs)
        self.message = '\n'.join([
            '*****************************',
            '  Sorry, there was an error registering ad.',
            '  The file ad.html is required to be in the templates folder',
            '  of your project so that the ad can be served.'])


class AdHtmlTooLarge(AmtServicesWrapperError):
    def __init__(self, size_of_ad, *args, **kwargs):
        super(AdHtmlTooLarge, self).__init__(*args, **kwargs)
        self.message = '\n'.join([
            '*****************************',
            '  Sorry, there was an error registering the ad.',
            '  Your local ad.html is {} bytes, but the maximum'.format(size_of_ad),
            '  template size uploadable to the ad server is',
            '  1048576 bytes.'])
                                 
#################
# API Exceptions
#################


class APIException(PsiturkException):
    pass


class EphemeralContainerDBError(PsiturkException):
    def __init__(self, db_address, *args, **kwargs):
        super(EphemeralContainerDBError, self).__init__(**kwargs)
        self.message = '\n'.join([
            f'The database URL you gave is {db_address}. This is unlikely to work ',
            'in an ephemeral container like Heroku either because it is run on a localhost or because it uses SQLite ',
            '(see https://devcenter.heroku.com/articles/sqlite3 for why that might not be a good idea). One better '
            'option might be to use a Heroku PostGRES table. Instructions to do that can be found here: ',
            'https://psiturk.readthedocs.io/en/latest/heroku.html'])


class HerokuCmdNotFound(PsiturkException):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.message = 'command `heroku` command not found. Please first install the heroku cmd cli.'


class HerokuNotLoggedIn(PsiturkException):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.message = 'Heroku cli not logged in. Please run `heroku auth:login` first.'


class HerokuNotAGitRepo(PsiturkException):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.message = 'Current directory is not a git repository. See https://devcenter.heroku.com/articles/git for help.'


class HerokuAppNotSet(PsiturkException):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.message = 'Heroku app not known. '



from __future__ import generator_stop
from psiturk.amt_services_wrapper import MTurkServicesWrapper

SESSION_SERVICES_MANAGER_MODE_KEY = 'services_manager_mode'


class PsiturkServicesManager:
    _cached_amt_services_wrapper = None

    @property
    def amt_services_wrapper(self):
        if not self._cached_amt_services_wrapper:
            self._cached_amt_services_wrapper = MTurkServicesWrapper()
        return self._cached_amt_services_wrapper

    @property
    def config(self):
        return self.amt_services_wrapper.config

    @property
    def mode(self):
        return self.amt_services_wrapper.get_mode().data

    @mode.setter
    def mode(self, value):
        result = self.amt_services_wrapper.set_mode(value)
        if not result.success:
            raise Exception(result.exception)

    @property
    def codeversion(self):
        return self.config['Task Parameters']['experiment_code_version']

    @property
    def amt_balance(self):
        return self.amt_services_wrapper.amt_balance().data


psiturk_services_manager = PsiturkServicesManager()



from __future__ import generator_stop
from flask import Blueprint, jsonify, make_response, request, current_app as app
from flask.json import JSONEncoder
from flask_restful import Api, Resource
from psiturk.dashboard import login_required
from psiturk.services_manager import psiturk_services_manager as services_manager
from psiturk.models import Participant, Campaign
from psiturk.experiment import app
from psiturk.psiturk_exceptions import *
from psiturk.amt_services_wrapper import WrapperResponse
from psiturk.amt_services import MTurkHIT
from psiturk.db import db_session
from psiturk.models import Base as DBModel
from apscheduler.job import Job
from apscheduler.triggers.base import BaseTrigger
import datetime
import pytz
from pytz.tzinfo import BaseTzInfo

api_blueprint = Blueprint('api', __name__, url_prefix='/api')
api = Api()


@api_blueprint.errorhandler(Exception)
def handle_exception(exception):
    message = exception.message if (hasattr(exception, 'message') and
                                    exception.message) else str(exception)
    # raise exception
    return jsonify({
        'exception': type(exception).__name__,
        'message': message
    }), 400


class PsiturkJSONEncoder(JSONEncoder):  # flask's jsonencoder class
    def default(self, obj):
        if isinstance(obj, (PsiturkException, WrapperResponse)):
            return obj.to_dict()

        if isinstance(obj, Exception):
            return {
                'exception': type(obj).__name__,
                'message': str(obj)
            }

        if isinstance(obj, datetime.timedelta):
            return str(obj)

        if isinstance(obj, MTurkHIT):
            return obj.__dict__

        if isinstance(obj, DBModel):
            return obj.object_as_dict()

        if isinstance(obj, (Job, BaseTrigger)):
            return obj.__getstate__()

        if isinstance(obj, BaseTzInfo):
            return str(obj)

        return JSONEncoder.default(self, obj)


api_blueprint.json_encoder = PsiturkJSONEncoder


@api.representation('application/json')
def output_json(data, code, headers=None):
    resp = make_response(jsonify(data), code)
    resp.headers.extend(headers or {})
    return resp


@api_blueprint.before_request
@login_required
def before_request():
    pass


class ServicesManager(Resource):
    def get(self):
        _return = {
            'mode': 'unavailable',
            'codeversion': 'unavailable',
            'amt_balance': 'unavailable',
            'aws_access_key_id': 'unavailable'
        }
        try:
            _return['mode'] = services_manager.mode
            _return['codeversion'] = services_manager.codeversion
            _return['amt_balance'] = services_manager.amt_balance
            _return['aws_access_key_id'] = services_manager.config.get('AWS Access',
                                                                       'aws_access_key_id')
        except PsiturkException:
            pass
        return _return


class AssignmentList(Resource):
    def get(self, assignment_id=None):
        all_workers = [worker.object_as_dict(filter_these=['datastring']) for worker in
                       Participant.query.filter(Participant.mode != 'debug').all()]
        return all_workers


class AssignmentsAction(Resource):
    def post(self, action=None):
        data = request.json
        if action == 'approve_all':
            response = services_manager.amt_services_wrapper.approve_all_assignments()
            if not response.success:
                raise response.exception
            return response.data['results']
        elif action == 'bonus_all':
            if 'reason' not in data or not data['reason']:
                raise APIException(message='bonus reason is missing!')

            response = services_manager.amt_services_wrapper.bonus_all_local_assignments(
                amount='auto', reason=data['reason'])
            if not response.success:
                raise response.exception
            return response.data['results'], 201
        else:
            raise APIException(message='action `{}` not recognized!'.format(action))


class Hits(Resource):
    def patch(self, hit_id):
        data = request.json
        if 'is_expired' in data and data['is_expired']:
            response = services_manager.amt_services_wrapper.expire_hit(hit_id)
            if not response.success:
                raise response.exception
        if 'action' in data:
            if data['action'] == 'approve_all':
                response = services_manager.amt_services_wrapper.approve_assignments_for_hit(hit_id)
                if not response.success:
                    raise response.exception
        return services_manager.amt_services_wrapper.get_hit(hit_id).data, 201

    def delete(self, hit_id):
        response = services_manager.amt_services_wrapper.delete_hit(hit_id)
        if not response.success:
            raise response.exception
        return '', 204


class HitList(Resource):
    def get(self, status=None):
        if status == 'active':
            hits = services_manager.amt_services_wrapper.get_active_hits().data
        else:
            hits = services_manager.amt_services_wrapper.get_all_hits().data
        hits = [hit.__dict__ for hit in hits]
        return hits


class HitsAction(Resource):
    def get(self, action=None):
        if action == 'expire_all':
            response = services_manager.amt_services_wrapper.expire_all_hits()
            if response.success:
                return response.data['results'], 201
            else:
                raise response.exception

        elif action == 'delete_all':
            response = services_manager.amt_services_wrapper.delete_all_hits()
            if response.success:
                return response.data['results'], 201
            else:
                raise response.exception

        elif action == 'approve_all':
            hits = services_manager.amt_services_wrapper.get_all_hits().data

            _return = []
            for hit in hits:
                hit_response = services_manager.amt_services_wrapper.approve_assignments_for_hit(
                    hit.options['hitid'], all_studies=True)
                _return.append(hit_response)
            return _return, 201

        else:
            raise APIException(message='action `{}` not recognized!'.format(action))


class Campaigns(Resource):
    def get(self, campaign_id):
        campaign = Campaign.query.filter(Campaign.id == campaign_id).one()
        return campaign

    def patch(self, campaign_id):
        campaign = Campaign.query.filter(Campaign.id == campaign_id).one()
        data = request.json
        did_something = False
        if 'is_active' in data and campaign.is_active and not data['is_active']:
            campaign.end()
            did_something = True
        elif 'goal' in data:
            goal = data['goal']

            completed_count = Participant.count_completed(
                codeversion=services_manager.codeversion,
                mode=services_manager.mode)

            assert goal > completed_count, 'Goal {} must be greater than current completed {}.'.format(
                goal, completed_count)

            campaign.goal = goal
            did_something = True

        if did_something:
            db_session.add(campaign)
            db_session.commit()

        return campaign


class CampaignList(Resource):
    def get(self):
        campaigns = Campaign.query.filter(Campaign.codeversion == services_manager.codeversion,
                                          Campaign.mode == services_manager.mode).all()
        return campaigns

    def post(self):
        data = request.json
        # check if one already exists that is active...
        if Campaign.active_campaign_exists():
            raise APIException(
                message='Active campaign already exists. Cancel that campaign first in order to create a new one.')
        codeversion = services_manager.codeversion
        mode = services_manager.mode
        campaign = Campaign.launch_new_campaign(codeversion=codeversion, mode=mode, **data)
        return campaign, 201


class Tasks(Resource):
    def delete(self, task_id):
        app.apscheduler.remove_job(str(task_id))
        return '', 204


class TaskList(Resource):
    def get(self):
        tasks = app.apscheduler.get_jobs()
        return tasks

    def post(self):
        data = request.json
        mode = services_manager.mode
        if data['name'] == 'approve_all':
            # check if we already have an 'approve_all'
            if app.apscheduler.get_job('approve_all'):
                raise APIException(message='`approve_all` job already exists')

            from psiturk.tasks import do_approve_all

            job = app.apscheduler.add_job(
                id='approve_all',
                args=[mode],
                func=do_approve_all,
                trigger='interval',
                minutes=max(int(float(data['interval']) * 60), 30),
                next_run_time=datetime.datetime.now(pytz.utc)
            )
            return job, 201

        raise APIException(message='task name `{}` not recognized!'.format(data['name']))


api.add_resource(ServicesManager, '/services_manager', '/services_manager/')

api.add_resource(AssignmentList, '/assignments', '/assignments/')
api.add_resource(AssignmentsAction, '/assignments/action/<action>')

api.add_resource(Hits, '/hit/<hit_id>')
api.add_resource(HitList, '/hits/', '/hits/<status>')
api.add_resource(HitsAction, '/hits/action/<action>')

api.add_resource(CampaignList, '/campaigns', '/campaigns/')
api.add_resource(Campaigns, '/campaigns/<campaign_id>')

api.add_resource(TaskList, '/tasks', '/tasks/')
api.add_resource(Tasks, '/tasks/<task_id>')

api.init_app(api_blueprint)



""" This module supports commandline functionality """
from __future__ import generator_stop
import argparse
import sys
import os
from psiturk.version import version_number


def process():
    """Figure out how we were invoked."""
    invoked_as = os.path.basename(sys.argv[0])

    if invoked_as == "psiturk":
        launch_shell()
    elif invoked_as == "psiturk-server":
        launch_server()
    elif invoked_as == "psiturk-shell":
        launch_shell()
    elif invoked_as == "psiturk-setup-example":
        setup_example()
    elif invoked_as == "psiturk-heroku-config":
        from psiturk.do_heroku_setup import do_heroku_setup
        do_heroku_setup()


def setup_example():
    """Add commands for testing, etc."""
    parser = argparse.ArgumentParser(
        description='Creates a simple default project (stroop) in the current\
        directory with the necessary psiTurk files.'
    )

    # Optional flags
    parser.add_argument(
        '-v', '--version', help='Print version number.', action="store_true"
    )
    args = parser.parse_args()

    # If requested version just print and quit
    if args.version:
        print(version_number)
    else:
        import psiturk.setup_example as se
        se.setup_example()


def launch_server():
    """Add commands for testing, etc."""
    parser = argparse.ArgumentParser(
        description='Launch psiTurk experiment webserver process on the\
        host/port defined in config.txt.'
    )

    # Optional flags
    parser.add_argument(
        '-v', '--version', help='Print version number.', action="store_true"
    )
    args = parser.parse_args()

    # If requested version just print and quit
    if args.version:
        print(version_number)
    else:
        import psiturk.experiment_server as es
        es.launch()


def launch_shell():
    """Add commands for testing, etc."""
    parser = argparse.ArgumentParser(
        description='Launch the psiTurk interactive shell.'
    )

    # Optional flags
    parser.add_argument(
        '-v', '--version', help='Print version number.', action="store_true"
    )

    script_group = parser.add_mutually_exclusive_group()
    script_group.add_argument(
        '-s', '--script', help='Run commands from a script file'
    )
    script_group.add_argument(
        '-e', '--execute', help='Execute one command specified on the command line'
    )
    script_group.add_argument(
        '-t', '--test', help='Run cmd2 unittest using provided file'
    )
    args, unknownargs = parser.parse_known_args()
    # If requested version just print and quit
    if args.version:
        print(version_number)
    else:
        import psiturk.psiturk_shell as ps
        if args.script:
            ps.run(script=args.script, quiet=True)
        elif args.test:
            ps.run(testfile=args.test, quiet=True)
        elif args.execute or unknownargs:
            if unknownargs:
                import shlex
                execute = ' '.join([shlex.quote(e) for e in unknownargs])
            else:
                execute = args.execute
            ps.run(execute=execute, quiet=True)
        else:
            ps.run()



# coding: utf-8
""" PsiturkShell is a commandline interface for psiTurk, which provides
functionality for maintaining the experiment server and interacting with
Mechanical Turk."""
from __future__ import generator_stop
try:
    from urllib.parse import quote_plus
except ImportError:
    from urllib import quote_plus
import urllib3.contrib.pyopenssl
from psiturk.utils import *
from psiturk.models import Participant
from psiturk import experiment_server_controller as control
from psiturk.psiturk_config import PsiturkConfig
from psiturk.version import version_number
from psiturk.amt_services_wrapper import MTurkServicesWrapper
from .psiturk_exceptions import *
import webbrowser
from docopt import docopt, DocoptExit
from cmd2 import Cmd
from fuzzywuzzy import process
import certifi
import random
import string
import os
import json
import time
import re
import subprocess
import sys
from functools import wraps
import shlex
urllib3.contrib.pyopenssl.inject_into_urllib3()
http = urllib3.PoolManager(
    cert_reqs='CERT_REQUIRED',
    ca_certs=certifi.where())


try:
    # prefer gnureadline if user has installed it
    import gnureadline as readline
except ImportError:
    try:
        # next pyreadline if user has installed it (windows)
        import pyreadline as readline
    except ImportError:
        import readline


def docopt_cmd(func):
    """
    This decorator is used to simplify the try/except block and pass the result
    of the docopt parsing to the called action.
    """
    @wraps(func)
    def helper_fn(self, arg):
        """helper function for docopt"""
        try:
            arg = shlex.split(arg)
            opt = docopt(helper_fn.__doc__, arg)
        except DocoptExit as exception:
            # The DocoptExit is thrown when the args do not match.
            # We print a message to the user and the usage block.
            self.poutput('Invalid Command!')
            self.poutput(exception)
            return
        except SystemExit:
            # The SystemExit exception prints the usage for --help
            # We do not need to do the print here.
            return
        return func(self, opt)

    return helper_fn


class PsiturkNetworkShell(Cmd, object):
    """ Creates the Psiturk Network Shell interfrace """

    _cached_amt_services_wrapper = None

    @property
    def amt_services_wrapper(self):
        if not self._cached_amt_services_wrapper:
            try:
                _wrapper = MTurkServicesWrapper(
                    config=self.config, mode=self.mode)
                self._cached_amt_services_wrapper = _wrapper
            except AmtServicesException as e:
                still_can_do = '\n'.join([
                    '',
                    'You can still use the psiturk server by running non-AWS commands such as:',
                    '- `psiturk server <subcommand>`',
                    '- `psiturk server on`',
                    '- `psiturk server off`',
                    '- `psiturk debug -p`'])
                message = e.message
                if not self.quiet:
                    message = '{}{}'.format(message, still_can_do)
                self.perror(message)
                sys.exit()
            except PsiturkException as e:
                self.poutput(e)

        return self._cached_amt_services_wrapper

    def postcmd(self, *args):
        if not self.quiet:
            self.prompt = self.color_prompt()
        return Cmd.postcmd(self, *args)

    def complete(self, text, state):
        """ Add space after a completion, makes tab completion with
        multi-word commands cleaner. """
        return Cmd.complete(self, text, state) + ' '

    def default(self, statement):
        cmd = statement.command

        ''' Collect incorrect and mistyped commands '''
        choices = ["help", "mode", "psiturk_status", "server", "shortcuts",
                   "worker", "db", "edit", "open", "config", "show",
                   "debug", "setup_example", "status", "amt_balance",
                   "download_datafiles", "exit", "hit", "load", "quit", "save",
                   "shell", "version"]
        self.poutput("{} is not a psiTurk command. See 'help'.".format(cmd))
        guess = process.extractOne(cmd, choices)
        if guess and guess[1] > 50:
            self.poutput("Did you mean this?\n\t{}".format(guess[0]))

    def __init__(self, config, server, mode='sandbox', quiet=False):
        persistent_history_file = config.get('Shell Parameters',
                                             'persistent_history_file')

        # Prevents running of commands by abbreviation
        self.abbrev = False
        self.debug = True
        self.help_path = os.path.join(os.path.dirname(__file__), "shell_help/")
        self.psiturk_header = 'psiTurk command help:'
        self.super_header = 'basic CMD command help:'

        self.config = config
        self.server = server

        self.mode = mode
        self.sandbox_hits = 0
        self.live_hits = 0

        Cmd.__init__(self, persistent_history_file=persistent_history_file)
        self.quiet = quiet
        self.feedback_to_output = True

        if not self.quiet and not self.amt_services_wrapper:
            sys.exit()

        if not self.quiet:
            self.maybe_update_hit_tally()
            self.prompt = self.color_prompt()
            self.intro = self.get_intro_prompt()
        else:
            self.intro = ''

    def server_off(self):
        if (self.server.is_server_running() == 'yes' or
                self.server.is_server_running() == 'maybe'):
            self.server.shutdown()
            self.poutput('Please wait. This could take a few seconds.')
            while self.server.is_server_running() != 'no':
                time.sleep(0.5)
        else:
            self.poutput('Your server is already off.')

    # +-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.
    #   basic command line functions
    # +-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.

    def get_intro_prompt(self):
        """ Overloads intro prompt with network-aware version if you can reach
        psiTurk.org, request system status message. """
        status_msg_url = 'https://raw.githubusercontent.com/NYUCCL/psiTurk/master/status_msg.txt'
        r = http.request('GET', status_msg_url)
        status_message = r.data.decode('utf-8')

        return status_message + colorize('psiTurk version ' + version_number +
                                         '\nType "help" for more information.',
                                         'green', False)

    def color_prompt(self):
        prompt = '[' + colorize('psiTurk', 'bold')
        server_string = ''
        server_status = self.server.is_server_running()
        if server_status == 'yes':
            server_string = colorize('on', 'green')
        elif server_status == 'no':
            server_string = colorize('off', 'red')
        elif server_status == 'maybe':
            server_string = colorize('status unknown', 'yellow')
        elif server_status == 'blocked':
            server_string = colorize('blocked', 'red')
        prompt += ' server:' + server_string
        mode_str = 'sdbx' if self.mode == 'sandbox' else 'live'
        prompt += ' mode:' + colorize(mode_str, 'bold')
        hit_count = self.sandbox_hits if self.mode == 'sandbox' else self.live_hits
        prompt += ' #HITs:' + str(hit_count)
        prompt += ']$ '
        return prompt

    # +-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.
    #   basic command line functions
    # +-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.

    def do_version(self, _):
        """ Print version number """
        self.poutput('psiTurk version ' + version_number)

    @docopt_cmd
    def do_config(self, arg):
        """
        Usage:
          config print
          config reload
          config help
        """
        if arg['print']:
            self.print_config(arg)
        elif arg['reload']:
            self.reload_config(arg)
        else:
            self.help_config()

    def complete_config(self, text, line, begidx, endidx):
        """ Tab-complete config command """
        config_commands = ('print', 'reload', 'help')
        return [i for i in config_commands if i.startswith(text)]

    def help_config(self):
        """ Help for config """
        with open(self.help_path + 'config.txt', 'r') as help_text:
            self.poutput(help_text.read())

    def print_config(self, _):
        """ Print configuration. """
        for section in self.config.sections():
            self.poutput('[%s]' % section)
            items = dict(self.config.items(section))
            for k in items:
                self.poutput("%(a)s=%(b)s" % {'a': k, 'b': items[k]})
            print('')

    def reload_config(self, _):
        """ Reload config. """
        # TODO: I don't think that this follows all the way through!
        restart_server = False
        if (self.server.is_server_running() == 'yes' or
                self.server.is_server_running() == 'maybe'):
            user_input = input("Reloading configuration requires the server "
                               "to restart. Really reload? y or n: ")
            if user_input != 'y':
                return
            restart_server = True
        self.config.load_config()
        if restart_server:
            self.server_restart()

    def do_setup_example(self, _):
        """ Load psiTurk demo."""
        from . import setup_example as se
        se.setup_example()

    def help_debug(self):
        """ Help for debug """
        with open(self.help_path + 'debug.txt', 'r') as help_text:
            self.poutput(help_text.read())

    @docopt_cmd
    def do_open(self, arg):
        """
        Usage: open
               open <folder>

        Opens folder or current directory using the local system's shell
        command 'open'.
        """
        if arg['<folder>'] is None:
            subprocess.call(["open"])
        else:
            subprocess.call(["open", arg['<folder>']])

    def do_eof(self, arg):
        """ Execute on EOF """
        return self.do_quit(arg)

    def do_exit(self, arg):
        """ Execute on exit """
        return self.do_quit(arg)

    def do_quit(self, _):
        """ Execute on quit """
        if (self.server.is_server_running() == 'yes' or
                self.server.is_server_running() == 'maybe'):
            user_input = input("Quitting shell will shut down experiment "
                               "server.  Really quit? y or n: ")
            if user_input == 'y':
                self.server_off()
            else:
                return
        return True

    def do_psiturk_status(self, _):
        """ Print psiTurk news """
        self.poutput(self.get_intro_prompt())

    # +-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.
    #   Local SQL database commands
    # +-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.

    def do_download_datafiles(self, _):
        """ Download datafiles. """
        contents = {"trialdata": lambda p: p.get_trial_data(), "eventdata":
                    lambda p: p.get_event_data(), "questiondata": lambda p:
                    p.get_question_data()}
        query = Participant.query.all()
        for k in contents:
            ret = "".join([contents[k](p) for p in query])
            temp_file = open(k + '.csv', 'w')
            temp_file.write(ret)
            temp_file.close()

    # +-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.
    #   hit management
    # +-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.
    def hit_list(self, active_hits, reviewable_hits, all_studies):
        """ List hits. """
        if active_hits:
            hits_data = (self.amt_services_wrapper.get_active_hits(
                all_studies)).data
        elif reviewable_hits:
            hits_data = (self.amt_services_wrapper.get_reviewable_hits(
                all_studies)).data
        else:
            hits_data = (self.amt_services_wrapper.get_all_hits(
                all_studies)).data
        if not hits_data:
            self.poutput('*** no hits retrieved')
        else:
            for hit in hits_data:
                self.poutput(hit)

    def _estimate_expenses(self, num_workers, reward):
        """ Returns tuple describing expenses:
        amount paid to workers
        amount paid to amazon"""

        # fee structure changed 07.22.15:
        # 20% for HITS with < 10 assignments
        # 40% for HITS with >= 10 assignments
        commission = 0.2
        if float(num_workers) >= 10:
            commission = 0.4
        work = float(num_workers) * float(reward)
        fee = work * commission
        return work, fee, work + fee

    def _confirm_dialog(self, prompt):
        """ Prompts for a 'yes' or 'no' to given prompt. """
        response = input(prompt).strip().lower()
        valid = {'y': True, 'ye': True, 'yes': True, 'n': False, 'no': False}
        while True:
            try:
                return valid[response]
            except KeyError:
                response = input("Please respond 'y' or 'n': ").strip().lower()

    def hit_create(self, num_workers, reward, duration,
                   require_qualification_ids=None, block_qualification_ids=None, **kwargs):

        # backwards compatibility
        if 'whitelist_qualification_ids' in kwargs and not require_qualification_ids:
            require_qualification_ids = kwargs['whitelist_qualification_ids']
        if 'blacklist_qualification_ids' in kwargs and not block_qualification_ids:
            block_qualification_ids = kwargs['blacklist_qualification_ids']

        if require_qualification_ids is None:
            require_qualification_ids = []
        if block_qualification_ids is None:
            block_qualification_ids = []

        # Argument retrieval and validation
        if num_workers is None:
            num_workers = input('number of participants? ').strip()
        try:
            num_workers = int(num_workers)
        except ValueError:
            self.poutput('*** number of participants must be a whole number')
            return
        if num_workers <= 0:
            self.poutput('*** number of participants must be greater than 0')
            return

        if reward is None:
            reward = input('reward per HIT? ').strip()
        p = re.compile(r'^\d*\.\d\d$')
        m = p.match(reward)
        if m is None:
            self.poutput('*** reward must have format [dollars].[cents]')
            return
        try:
            reward = float(reward)
        except (ValueError, OverflowError):
            self.poutput('*** reward must be in format [dollars].[cents]')
            return

        if duration is None:
            duration = input(
                'duration of hit (in hours, it can be decimals)? ').strip()
        try:
            duration = float(duration)
        except ValueError:
            self.poutput('*** duration must a number')
            return
        if duration <= 0:
            self.poutput('*** duration must be greater than 0')
            return

        _, fee, total = self._estimate_expenses(num_workers, reward)

        if not self.quiet:
            dialog_query = '\n'.join(['*****************************',
                                      '    Max workers: %d' % num_workers,
                                      '    Reward: $%.2f' % reward,
                                      '    Duration: %s hours' % duration,
                                      '    Fee: $%.2f' % fee,
                                      '    ________________________',
                                      '    Total: $%.2f' % total,
                                      'Create %s HIT [y/n]? ' % colorize(self.mode, 'bold')])
            if not self._confirm_dialog(dialog_query):
                self.poutput('*** Cancelling HIT creation.')
                return

        try:
            create_hit_response = self.amt_services_wrapper.create_hit(num_workers=num_workers, reward=reward,
                                                                       duration=duration,
                                                                       require_qualification_ids=require_qualification_ids,
                                                                       block_qualification_ids=block_qualification_ids)

            if create_hit_response.status != 'success':
                self.poutput('Error during hit creation.')
                print(create_hit_response)
                return
            else:
                self.maybe_update_hit_tally()

            hit_id = create_hit_response.data['hit_id']

            # TODO: "with qualification type ids..."
            self.poutput('\n'.join(['*****************************',
                                    '  Created %s HIT' % colorize(
                                        self.mode, 'bold'),
                                    '    HITid: %s' % str(hit_id),
                                    '    Max workers: %d' % num_workers,
                                    '    Reward: $%.2f' % reward,
                                    '    Duration: %s hours' % duration,
                                    '    Fee: $%.2f' % fee,
                                    '    ________________________',
                                    '    Total: $%.2f' % total]))

            # Print the Ad Url
            base = self.config.get_ad_url()
            assignmentid = str(self.random_id_generator())
            hitid = str(self.random_id_generator())
            workerid = str(self.random_id_generator())

            ad_url = f'{base}?mode={self.mode}&assignmentId=debug{assignmentid}&hitId=debug{hitid}&workerId=debug{workerid}'
            self.poutput(f'  Ad URL: {ad_url}')
            self.poutput(
                "Note: This url cannot be used to run your full psiTurk experiment."
                "It is only for testing your ad."
            )

            # Print the Mturk Url
            mturk_url = ''
            if self.mode == 'sandbox':
                mturk_url_base = 'https://workersandbox.mturk.com'
            else:  # self.mode == 'live':
                mturk_url_base = 'https://worker.mturk.com'
            title = quote_plus(
                str(self.config.get('HIT Configuration', 'title', raw=True)))
            mturk_url = f'{mturk_url_base}/projects?filters%5Bsearch_term%5D={title}'

            self.poutput(f'  MTurk URL: {mturk_url}')
            self.poutput(
                "Hint: In OSX, you can open a terminal link using cmd + click")
        except Exception as e:
            self.poutput(e)

    # +-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.
    #   worker management
    # +-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.
    def worker_list(self, chosen_hits=None, status=None, all_studies=False):
        try:
            workers = (self.amt_services_wrapper.get_assignments(
                hit_ids=chosen_hits, assignment_status=status, all_studies=all_studies)).data['assignments']
            if not len(workers):
                self.poutput("*** no workers match your request")
            else:
                worker_json = json.dumps(workers, indent=4,
                                         separators=(',', ': '), default=str)
                if worker_json:
                    self.poutput(worker_json)

        except Exception as e:
            self.poutput(colorize(repr(e), 'red'))

    # TODO: consider renaming all to all_hits; shadows built-in
    def worker_approve(self, all=False, chosen_hits=None, assignment_ids=None, all_studies=False, force=False):

        if all_studies and not force:
            self.poutput(
                'option --all-studies must be used along with --force')
            return

        all_studies_msg = ' for the current study'
        if all_studies:
            all_studies_msg = ' from all studies'

        force_msg = ''
        if force:
            force_msg = " even if they're not found in the local psiturk db"

        if all:
            self.poutput("Approving all submissions{}{}...".format(
                all_studies_msg, force_msg))
            result = self.amt_services_wrapper.approve_all_assignments(
                all_studies=all_studies)
            if not result.success:
                return self.poutput(result)
            for _result in result.data['results']:
                self.poutput(_result)
        elif chosen_hits:
            self.poutput("Approving submissions for HITs {}{}{}".format(
                ' '.join(chosen_hits), all_studies_msg, force_msg))
            for hit_id in chosen_hits:
                result = self.amt_services_wrapper.approve_assignments_for_hit(
                    hit_id, all_studies=all_studies)
                if not result.success:
                    return self.poutput(result)
                for _result in result.data['results']:
                    self.poutput(_result)
        else:
            self.poutput("Approving specified submissions{}{}...".format(
                all_studies_msg, force_msg))
            for assignment_id in assignment_ids:
                result = self.amt_services_wrapper.approve_assignment_by_assignment_id(
                    assignment_id, all_studies=all_studies)
                self.poutput(result)

    # +-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.
    #   server management
    # +-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.

    @docopt_cmd
    def do_server(self, arg):
        """
        Usage:
          server on
          server off
          server restart
          server log
          server help
        """
        if arg['on']:
            self.server_on()
        elif arg['off']:
            self.server_off()
        elif arg['restart']:
            self.server_restart()
        elif arg['log']:
            self.server_log()
        else:
            self.help_server()

    def complete_server(self, text, line, begidx, endidx):
        """ Tab-complete server command """
        server_commands = ('on', 'off', 'restart', 'log', 'help')
        return [i for i in server_commands if i.startswith(text)]

    def help_server(self):
        """ Help for server """
        with open(self.help_path + 'server.txt', 'r') as help_text:
            self.poutput(help_text.read())

    def server_on(self):
        """ Start experiment server """
        self.server.startup()
        time.sleep(0.5)

    def server_restart(self):
        """ Restart experiment server """
        self.server_off()
        self.server_on()

    def server_log(self):
        """ Launch log """
        logfilename = self.config.get('Server Parameters', 'logfile')
        if sys.platform == "darwin":
            args = ["open", "-a", "Console.app", logfilename]
        else:
            args = ["xterm", "-e", "'tail -f %s'" % logfilename]
        subprocess.Popen(args, close_fds=True)
        self.poutput("Log program launching...")

    def do_status(self, arg):  # overloads do_status with AMT info
        """ Notify user of server status. """
        server_status = self.server.is_server_running()
        if server_status == 'yes':
            self.poutput('Server: ' + colorize('currently online', 'green'))
        elif server_status == 'no':
            self.poutput('Server: ' + colorize('currently offline', 'red'))
        elif server_status == 'maybe':
            self.poutput('Server: ' + colorize('status unknown', 'yellow'))
        elif server_status == 'blocked':
            self.poutput('Server: ' + colorize('blocked', 'red'))

        # server_status = self.server.is_server_running()  # Not used
        self.update_hit_tally()
        hit_count = self.sandbox_hits if self.mode == 'sandbox' else self.live_hits
        self.poutput('AMT worker site - ' + colorize(self.mode, 'bold') + ': '
                     + str(hit_count) + ' HITs available')

    def maybe_update_hit_tally(self):
        if not self.quiet:
            self.update_hit_tally()

    def update_hit_tally(self):
        """ Tally hits """
        num_hits = (self.amt_services_wrapper.tally_hits()).data['hit_tally']
        if self.mode == 'sandbox':
            self.sandbox_hits = num_hits
        else:
            self.live_hits = num_hits

    # +-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.
    #   hit management
    # +-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.
    def do_amt_balance(self, _):
        """ Get MTurk balance """
        balance = (self.amt_services_wrapper.amt_balance()).data
        self.poutput(balance)

    def help_amt_balance(self):
        """ Get help for amt_balance. """
        with open(self.help_path + 'amt.txt', 'r') as help_text:
            self.poutput(help_text.read())

    # +-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.
    #   qualification type management
    # +-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.
    @docopt_cmd
    def do_qualifications(self, arg):
        """
        Usage: qualifications list
        """
        if arg['list']:
            self.do_list_qualifications()

    def do_list_qualifications(self):
        result = (self.amt_services_wrapper.list_qualification_types(MustBeOwnedByCaller=True))
        if not result.success:
            return self.poutput(result)
        qualification_types = result.data['qualification_types']
        only_these_keys = ['Name', 'Description', 'QualificationTypeId']
        q_type_projection = [{key: q_type[key] for key in only_these_keys} for q_type in qualification_types if q_type['QualificationTypeStatus'] == 'Active']
        if not q_type_projection:
            self.poutput('No custom qualifications found.')
        for q_type in q_type_projection:
            self.poutput(q_type)

    # +-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.
    #   Basic shell commands
    # +-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.+-+.
    @docopt_cmd
    def do_mode(self, arg):
        """
        Usage: mode
               mode <which>
        """
        if arg['<which>'] is None:
            # then toggle the `mode`
            mode = 'live' if self.mode == 'sandbox' else 'sandbox'
        else:
            # then accept what's given
            mode = arg['<which>']
        self.set_mode(mode)

    def help_mode(self):
        """ Help """
        with open(self.help_path + 'mode.txt', 'r') as help_text:
            self.poutput(help_text.read())

    def set_mode(self, mode):
        known_modes = ['sandbox', 'live']
        if mode not in known_modes:
            self.poutput('mode {} is not recognized.'.format(mode))
            return

        current_mode = self.mode
        if mode == current_mode:
            self.poutput('mode is already {}'.format(mode))
            return
        else:
            restart_server = False
            if self.server.is_server_running() == 'yes' or self.server.is_server_running() == 'maybe':
                if not self.quiet:
                    r = input("Switching modes requires the server to restart. Really "
                              "switch modes? y or n: ")
                    if r != 'y':
                        return
                    else:
                        restart_server = True

            self.mode = mode
            self.amt_services_wrapper.set_mode(mode)

            self.update_hit_tally()
            self.poutput('Entered %s mode' % colorize(mode, 'bold'))
            if restart_server:
                self.server_restart()

    @docopt_cmd
    def do_hit(self, arg):
        """
        Usage:
          hit create [<num_workers> <reward> <duration>] [--require-qualification-id <require_qualification_id>]... [--block-qualification-id <block_qualification_id>]...
          hit extend <HITid> [(--assignments <number>)] [(--expiration <minutes>)]
          hit expire (--all | <HITid> ...)
          hit delete (--all | <HITid> ...) [--all-studies]
          hit list [--active | --reviewable] [--all-studies]
          hit help
        """

        all_studies = arg['--all-studies']

        if arg['create']:
            self.hit_create(arg['<num_workers>'], arg['<reward>'], arg['<duration>'],
                            require_qualification_ids=arg['<require_qualification_id>'],
                            block_qualification_ids=arg['<block_qualification_id>'])
        elif arg['extend']:
            result = self.amt_services_wrapper.extend_hit(
                arg['<HITid>'][0], assignments=arg['<number>'], minutes=arg['<minutes>'])
            self.poutput(result)
        elif arg['expire']:
            did_something = False
            if arg['--all']:
                result = self.amt_services_wrapper.expire_all_hits()
                if not result.success:
                    return self.poutput(result)
                for _result in result.data['results']:
                    self.poutput(_result)
                did_something = True
            elif arg['<HITid>']:
                did_something = True
                for hit_id in arg['<HITid>']:
                    result = self.amt_services_wrapper.expire_hit(hit_id)
                    self.poutput(result)
            if did_something:
                self.update_hit_tally()
        elif arg['delete']:
            did_something = False
            if arg['--all']:
                result = self.amt_services_wrapper.delete_all_hits(all_studies=all_studies)
                if not result.success:
                    return self.poutput(result)
                results = result.data['results']
                for _result in results:
                    self.poutput(_result)
                did_something = True
            elif arg['<HITid>']:
                did_something = True
                for hit_id in arg['<HITid>']:
                    result = self.amt_services_wrapper.delete_hit(hit_id)
                    self.poutput(result)
            if did_something:
                self.update_hit_tally()
        elif arg['list']:
            self.hit_list(arg['--active'], arg['--reviewable'],
                          all_studies)
        else:
            self.help_hit()

    hit_commands = ('create', 'extend', 'expire', 'delete', 'list')

    def complete_hit(self, text, line, begidx, endidx):
        """ Tab-complete hit command. """
        return [i for i in PsiturkNetworkShell.hit_commands if
                i.startswith(text)]

    def help_hit(self):
        """ HIT help """
        with open(self.help_path + 'hit.txt', 'r') as help_text:
            self.poutput(help_text.read())

    @docopt_cmd
    def do_worker(self, arg):
        """
        Usage:
          worker approve (--all | --hit <hit_id> ... | <assignment_id> ...) [--all-studies] [--force]
          worker reject (--hit <hit_id> | <assignment_id> ...) [--all-studies]
          worker unreject (--hit <hit_id> | <assignment_id> ...) [--all-studies]
          worker bonus (--amount <amount> | --auto) [--reason=<reason>] (--all | --hit <hit_id> | <assignment_id> ...) [--override-bonused-status] [--all-studies]
          worker list [--submitted | --approved | --rejected] [(--hit <hit_id> ...)] [--all-studies]
          worker help

        Options:
          --reason REASON    the reason...
        """
        all_studies = arg['--all-studies']
        if arg['approve']:
            self.worker_approve(arg['--all'], arg['<hit_id>'],
                                arg['<assignment_id>'], all_studies, arg['--force'])
        elif arg['reject']:
            results = None
            if arg['<hit_id>']:
                result = self.amt_services_wrapper.reject_assignments_for_hit(
                    arg['<hit_id>'], all_studies=all_studies)
                if not result.success:
                    return self.poutput(result)
                results = result.data['results']
            elif arg['<assignment_id>']:
                result = self.amt_services_wrapper.reject_assignments(
                    arg['<assignment_id>'], all_studies=all_studies)
                if not result.success:
                    return self.poutput(result)
                results = result.data['results']
            if results:
                for _result in results:
                    self.poutput(_result)

        elif arg['unreject']:
            if arg['<hit_id>']:
                self.amt_services_wrapper.unreject_assignments_for_hit(
                    arg['<hit_id>'], all_studies=all_studies)
            elif arg['<assignment_id>']:
                for assignment_id in arg['<assignment_id>']:
                    result = self.amt_services_wrapper.unreject_assignment(
                        assignment_id, all_studies=all_studies)
                    self.poutput(result)

        elif arg['list']:
            status = None
            if arg['--submitted']:
                status = 'Submitted'

            elif arg['--approved']:
                status = 'Approved'

            elif arg['--rejected']:
                status = 'Rejected'

            self.worker_list(
                status=status, chosen_hits=arg['<hit_id>'], all_studies=all_studies)

        elif arg['bonus']:
            reason = arg['--reason']
            if isinstance(reason, list):
                reason = ' '.join(reason)
            if not reason:
                if self.config.get('Shell Parameters', 'bonus_message'):
                    reason = self.config.get(
                        'Shell Parameters', 'bonus_message')
                    self.poutput(
                        f'Using bonus `reason` from config file: "{reason}"')
                while not reason:
                    user_input = input("Type the reason for the bonus. Workers "
                                       "will see this message: ")
                    reason = user_input

            override_bonused_status = arg['--override-bonused-status']

            if arg['--auto']:
                amount = 'auto'
            else:
                amount = float(arg['<amount>'])

            results = None
            if arg['<hit_id>']:
                result = (self.amt_services_wrapper.bonus_assignments_for_hit(
                    arg['<hit_id>'][0], amount, reason, all_studies=all_studies, override_bonused_status=override_bonused_status))
                if not result.success:
                    return self.poutput(result)
                results = result.data['results']

            elif arg['--all']:
                result = self.amt_services_wrapper.bonus_all_local_assignments(
                    amount, reason, override_bonused_status)
                if not result.success:
                    return self.poutput(result)
                results = result.data['results']

            elif arg['<assignment_id>']:
                results = [
                    self.amt_services_wrapper.bonus_assignment_for_assignment_id(
                        assignment_id, amount, reason, override_bonused_status) for assignment_id in arg['<assignment_id>']]

            if results:
                [self.poutput(_result) for _result in results]
        else:
            self.help_worker()

    worker_commands = ('approve', 'reject', 'unreject',
                       'bonus', 'list', 'help')

    def complete_worker(self, text, line, begidx, endidx):
        """ Tab-complete worker command. """
        return [i for i in PsiturkNetworkShell.worker_commands if
                i.startswith(text)]

    def help_worker(self):
        """ Help for worker command. """
        with open(self.help_path + 'worker.txt', 'r') as help_text:
            self.poutput(help_text.read())

    @docopt_cmd
    def do_debug(self, arg):
        """
        Usage: debug [options]

        -p, --print-only        just provides the URL, doesn't attempt to
                                launch browser
        """
        try:
            base_url = self.config.get_ad_url()
            self.pfeedback('generating debug url using `ad_url` config var')
        except PsiturkException:
            self.pfeedback('`ad_url_*` config vars not set; using Server '
                           'Parameters host and port vars.')
            host = self.config.get('Server Parameters', 'host')
            port = self.config.get('Server Parameters', 'port')
            base_url = f"http://{host}:{port}/ad"

        launch_url = (f'{base_url}?'
                      f'assignmentId=debug{self.random_id_generator()}'
                      f'&hitId=debug{self.random_id_generator()}'
                      f'&workerId=debug{self.random_id_generator()}'
                      f'&mode=debug')

        if arg['--print-only']:
            self.poutput(launch_url)
        else:
            self.poutput("Launching browser pointed at your randomized debug link, "
                         "\n\t" + launch_url)
            webbrowser.open(launch_url, new=1, autoraise=True)

    def random_id_generator(self, size=6, chars=string.ascii_uppercase +
                            string.digits):
        """ Generate random id numbers """
        return ''.join(random.choice(chars) for x in range(size))

    # Modified version of standard cmd help which lists psiturk commands first.
    def do_help(self, arg):
        if arg:
            try:
                func = getattr(self, 'help_' + arg)
            except AttributeError:
                try:
                    doc = getattr(self, 'do_' + arg).__doc__
                    if doc:
                        self.stdout.write("%s\n" % str(doc))
                        return
                except AttributeError:
                    pass
                self.stdout.write("%s\n" % str(self.nohelp % (arg,)))
                return
            func()
        else:
            # Modifications start here
            names = dir(PsiturkNetworkShell)
            super_names = dir(Cmd)
            new_names = [m for m in names if m not in super_names]
            help_struct = {}
            cmds_psiturk = []
            cmds_super = []
            for name in names:
                if name[:5] == 'help_':
                    help_struct[name[5:]] = 1
            names.sort()
            prevname = ''
            for name in names:
                if name[:3] == 'do_':
                    if name == prevname:
                        continue
                    prevname = name
                    cmd = name[3:]
                    if cmd in help_struct:
                        del help_struct[cmd]
                    if name in new_names:
                        cmds_psiturk.append(cmd)
                    else:
                        cmds_super.append(cmd)
            self.stdout.write("%s\n" % str(self.doc_leader))
            self.print_topics(self.psiturk_header, cmds_psiturk, 15, 80)
            self.print_topics(self.misc_header, list(
                help_struct.keys()), 15, 80)
            self.print_topics(self.super_header, cmds_super, 15, 80)


def run(script=None, execute=None, testfile=None, quiet=False):
    try:
        using_libedit = 'libedit' in readline.__doc__
        if using_libedit:
            print('\n'.join([
                'libedit version of readline detected.',
                'readline will not be well behaved, which may cause all sorts',
                'of problems for the psiTurk shell. We highly recommend installing',
                'the gnu version of readline by running "sudo pip install gnureadline".',
                'Note: "pip install readline" will NOT work because of how the OSX',
                'pythonpath is structured.']))
    except TypeError:
        pass  # pyreadline doesn't have anything for __doc__
    # Drop arguments which were already processed in command_line.py
    sys.argv = [sys.argv[0]]
    config = PsiturkConfig()
    config.load_config()
    server = control.ExperimentServerController(config)
    shell = PsiturkNetworkShell(
        config, server,
        mode=config.get('Shell Parameters', 'launch_in_mode'),
        quiet=quiet)

    if script:
        shell.runcmds_plus_hooks([f'load {script}'])
    elif execute:
        shell.runcmds_plus_hooks([execute])
    else:
        shell.cmdloop()



# -*- coding: utf-8 -*-
""" This module sets up a demo experiment to use with psiTurk.  """
from __future__ import generator_stop
import os
from distutils import dir_util, file_util

EXAMPLE_DIR = os.path.join(os.path.dirname(__file__), "example")
DEFAULT_CONFIG_FILE = os.path.join(os.path.dirname(__file__),
                                   "default_configs/local_config_defaults.txt")
DEFAULT_GLOBAL_CONFIG_FILE = os.path.join(os.path.dirname(__file__),
                                          "default_configs/global_config_defaults.txt")

if 'PSITURK_GLOBAL_CONFIG_LOCATION' in os.environ:
    GLOBAL_CONFIG_PATH = os.path.join(os.environ['PSITURK_GLOBAL_CONFIG_LOCATION'],
                                      ".psiturkconfig")
else:
    GLOBAL_CONFIG_PATH = "~/.psiturkconfig"

GLOBAL_CONFIG_FILE = os.path.expanduser(GLOBAL_CONFIG_PATH)
EXAMPLE_TARGET = os.path.join(os.curdir, "psiturk-example")


def setup_example():
    """ Setup example """
    if os.path.exists(EXAMPLE_TARGET):
        print("Error, `psiturk-example` directory already exists. Please \
            remove it then re-run the command.")
    else:
        print("Creating new folder `psiturk-example` in the current working \
            directory")
        os.mkdir(EXAMPLE_TARGET)

        print("Copying", EXAMPLE_DIR, "to", EXAMPLE_TARGET)
        dir_util.copy_tree(EXAMPLE_DIR, EXAMPLE_TARGET)

        file_util.move_file(os.path.join(EXAMPLE_TARGET, 'config.txt.sample'), os.path.join(EXAMPLE_TARGET, 'config.txt'))
        file_util.move_file(os.path.join(EXAMPLE_TARGET, 'custom.py.sample'), os.path.join(EXAMPLE_TARGET, 'custom.py'))


        # change to target directory
        os.chdir(EXAMPLE_TARGET)


if __name__ == "__main__":
    setup_example()



from __future__ import generator_stop
import datetime
import io
import csv
import json
from sqlalchemy import Column, Integer, String, DateTime, Float, Text, Boolean, func, inspect
from sqlalchemy.orm import validates
from sqlalchemy.ext.declarative import declarative_base
from psiturk.db import db_session
from .psiturk_config import PsiturkConfig
from typing import List
from itertools import groupby
from .tasks import do_campaign_round
from apscheduler.jobstores.base import JobLookupError

config = PsiturkConfig()
config.load_config()


TABLENAME = config.get('Database Parameters', 'table_name')
CODE_VERSION = config.get('Task Parameters', 'experiment_code_version')

# Base class

Base = declarative_base()
Base.query = db_session.query_property()


def object_as_dict(self, filter_these: list = None):
    if filter_these is None:
        filter_these = []
    return {c.key: getattr(self, c.key) for c in
            inspect(self).mapper.column_attrs if c.key not in filter_these}


Base.object_as_dict = object_as_dict


class Participant(Base):
    """
    Object representation of a participant in the database.
    """
    __tablename__ = TABLENAME

    uniqueid = Column(String(128), primary_key=True)
    assignmentid = Column(String(128), nullable=False)
    workerid = Column(String(128), nullable=False)
    hitid = Column(String(128), nullable=False)
    ipaddress = Column(String(128))
    browser = Column(String(128))
    platform = Column(String(128))
    language = Column(String(128))
    cond = Column(Integer)
    counterbalance = Column(Integer)
    codeversion = Column(String(128))
    beginhit = Column(DateTime)
    beginexp = Column(DateTime)
    endhit = Column(DateTime)
    bonus = Column(Float, default=0)
    status = Column(Integer, default=1)
    mode = Column(String(128))
    if 'postgres://' in config.get('Database Parameters', 'database_url').lower():
        datastring = Column(Text)
    else:
        datastring = Column(Text(4294967295))

    def __init__(self, **kwargs):
        self.uniqueid = "{workerid}:{assignmentid}".format(**kwargs)
        self.status = 1
        self.codeversion = CODE_VERSION
        self.beginhit = datetime.datetime.now(datetime.timezone.utc)
        for key in kwargs:
            setattr(self, key, kwargs[key])

    def __repr__(self):
        return "Subject(uniqueid|%s, condition|%s, status|%s, codeversion|%s)" % (
            self.uniqueid,
            self.cond,
            self.status,
            self.codeversion)

    def get_trial_data(self):
        try:
            trialdata = json.loads(self.datastring)["data"]
        except (TypeError, ValueError):
            # There was no data to return.
            print(("No trial data found in record:", self))
            return ""

        try:
            with io.StringIO() as outstring:
                csvwriter = csv.writer(outstring)
                for trial in trialdata:
                    csvwriter.writerow(
                        (
                            self.uniqueid,
                            trial["current_trial"],
                            trial["dateTime"],
                            json.dumps(trial["trialdata"])
                         )
                    )
                return outstring.getvalue()
        except:
            print(("Error reading record:", self))
            return ""

    def get_event_data(self):
        try:
            eventdata = json.loads(self.datastring)["eventdata"]
        except (ValueError, TypeError):
            # There was no data to return.
            print(("No event data found in record:", self))
            return ""

        try:
            with io.StringIO() as outstring:
                csvwriter = csv.writer(outstring)
                for event in eventdata:
                    csvwriter.writerow(
                        (
                            self.uniqueid,
                            event["eventtype"],
                            event["interval"],
                            event["value"],
                            event["timestamp"]
                         )
                    )
                return outstring.getvalue()
        except:
            print(("Error reading record:", self))
            return ""

    def get_question_data(self):
        try:
            questiondata = json.loads(self.datastring)["questiondata"]
        except (TypeError, ValueError):
            # There was no data to return.
            print(("No question data found in record:", self))
            return ""

        try:
            with io.StringIO() as outstring:
                csvwriter = csv.writer(outstring)
                for question in questiondata:
                    csvwriter.writerow(
                        (
                            self.uniqueid,
                            question,
                            questiondata[question]
                         )
                    )
                return outstring.getvalue()
        except:
            print(("Error reading record:", self))

            return ""
            
    @classmethod
    def count_completed(cls, codeversion, mode):
        completed_statuses = [3, 4, 5, 7]
        return cls.query.with_entities(func.count()).filter(
                cls.status.in_(completed_statuses),
                cls.codeversion == codeversion,
                cls.mode == mode
            ).scalar()

    @classmethod
    def count_workers_grouped(cls, query=None, group_bys: List[str] = None):
        if group_bys is None:
            group_bys = ['codeversion', 'mode', 'status']
        group_by_labels = group_bys + ['count']
        group_bys = [getattr(cls, group_by) for group_by in group_bys]
        if not query:
            query = cls.query
        for group_by in group_bys:
            query = query.group_by(group_by).order_by(group_by.desc())
        entities = group_bys + [func.count()]

        query = query.with_entities(*entities)
        results = query.all()

        def list_to_grouped_dicts(results):
            parsed_results = {}
            for k, group in groupby(results, lambda row: row[0]):  # k will be codeversion
                group = list(group)
                if len(group[0]) > 2:
                    parsed_results[k] = list_to_grouped_dicts([row[1:] for row in group])
                else:
                    parsed_results.update({k: v for k, v in group})
            return parsed_results

        # TODO: Is this call actually doing anything? Looks like it is ignored.
        parsed_results = list_to_grouped_dicts(results)

        zipped_results = [dict(zip(group_by_labels, row)) for row in results]
        return zipped_results

    @classmethod
    def all_but_datastring(cls):
        query = cls.query
        query = query.with_entities(*[c for c in cls.__table__.c if c.name != 'datastring'])
        return query.all()


class Hit(Base):
    """
    """
    __tablename__ = 'amt_hit'
    hitid = Column(String(128), primary_key=True)


class Campaign(Base):
    """
    """
    __tablename__ = 'campaign'
    id = Column(Integer, primary_key=True)
    codeversion = Column(String(128), nullable=False)
    mode = Column(String(128), nullable=False)
    goal = Column(Integer, nullable=False)
    minutes_between_rounds = Column(Integer, nullable=False)
    assignments_per_round = Column(Integer, nullable=False)
    hit_reward = Column(Float, nullable=False)
    hit_duration_hours = Column(Float, nullable=False)
    is_active = Column(Boolean, default=True)
    created = Column(DateTime, default=datetime.datetime.utcnow)
    ended = Column(DateTime, default=None)

    @validates('goal', 'minutes_between_rounds', 'assignments_per_round',
               'hit_duration_hours')
    def validate_greater_than_zero(self, key, value):
        if key == 'goal':
            self._validate_greater_than_already_completed(value)
        assert value > 0, 'Property `{}` must be greater than 0'.format(key)
        return value

    @validates('hit_reward')
    def validate_hit_reward(self, key, hit_reward):
        assert hit_reward >= 0, \
            f'Hit reward must be greater than or equal to zero, got {hit_reward}'
        return hit_reward

    def _validate_greater_than_already_completed(self, goal):
        count_completed = Participant.count_completed(
            codeversion=self.codeversion, mode=self.mode)
        assert goal > count_completed, \
            f'Goal ({goal}) must be greater than the count of '\
            f'already-completed {count_completed}).'
    
    @validates('mode')
    def validate_mode(self, key, mode):
        assert mode in ['sandbox', 'live'], 'Mode {} not recognized.'.format(mode)
        return mode

    @validates('is_active')
    def validate_is_active(self, key, is_active):
        if is_active:
            assert not self.active_campaign_exists(), 'No no, there can be only one active campaign.'
        return is_active

    def __init__(self, **kwargs):
        self.codeversion = CODE_VERSION
        for key in kwargs:
            setattr(self, key, kwargs[key])

    @property
    def campaign_job_id(self):
        return 'campaign-{}'.format(self.id)

    def end(self):
        self.is_active = False
        self.ended = datetime.datetime.now(datetime.timezone.utc)
        from .experiment import app
        try:
            app.apscheduler.remove_job(self.campaign_job_id)
        except JobLookupError:
            pass
        return self

    @classmethod
    def active_campaign_exists(cls):
        subquery = cls.query.filter(cls.is_active.is_(True)).exists()
        query = db_session.query(subquery)
        _return = query.scalar()
        return _return

    @classmethod
    def launch_new_campaign(cls, **kwargs):
        kwargs['is_active'] = True
        new_campaign = cls(**kwargs)
        db_session.add(new_campaign)
        db_session.commit()

        _kwargs = {
            'campaign': new_campaign,
            'job_id': new_campaign.campaign_job_id
        }
        from .experiment import app
        app.apscheduler.add_job(
            id=new_campaign.campaign_job_id,
            func=do_campaign_round,
            kwargs=_kwargs,
            trigger='interval',
            minutes=new_campaign.minutes_between_rounds,
            next_run_time=datetime.datetime.now(datetime.timezone.utc)
        )

        return new_campaign



from __future__ import generator_stop
from gunicorn.app.base import Application
from gunicorn import util
import multiprocessing
from psiturk.psiturk_config import PsiturkConfig
import os
import sys
import hashlib
import signal
from gevent import monkey
monkey.patch_all()

config = PsiturkConfig()
config.load_config()


def sigint_handler(signal, frame):
    """
    Give feedback when ^C is submitted by user.

    Unfortunately, this does not tell psiturk interactive to update server status
    """
    print('^C: shutting down server processes.')
    sys.exit(0)


signal.signal(signal.SIGINT, sigint_handler)


class ExperimentServer(Application):
    """
    Custom Gunicorn Server Application that serves up the Experiment application
    """

    def __init__(self, app_dir=None):
        """__init__ method
        Load the base config and assign some core attributes.
        """

        self.loglevels = ["debug", "info", "warning", "error", "critical"]
        self.load_user_config()
        self.usage = None
        self.callable = None
        self.options = self.user_options
        self.prog = None
        self.do_load_config()
        if app_dir:
            self.cfg.set('chdir', app_dir) # get directory
            self.chdir() # change to it
        self.user_options = {}
        print("Now serving on", "http://" + self.options["bind"])

    def init(self, *args):
        """init method
        Takes our custom options from self.options and creates a config
        dict which specifies custom settings.
        """
        cfg = {}
        for k, v in list(self.options.items()):
            if k.lower() in self.cfg.settings and v is not None:
                cfg[k.lower()] = v
        return cfg

    def load(self):
        """load method
        Imports our application and returns it to be run.
        """
        return util.import_app("psiturk.experiment:app")

    def load_user_config(self):
        # config calls these threads to avoid confusing with workers
        workers = config.get('Server Parameters', "threads")
        if workers == "auto":
            workers = str(multiprocessing.cpu_count() * 2 + 1)

        if int(workers) > 1 and config.getboolean('Server Parameters',
                                                  'do_scheduler'):
            raise Exception((
                'Scheduler is not thread-safe, '
                'but {} gunicorn workers requested! Refusing to start!'
                ).format(workers)
            )

        self.loglevels = ["debug", "info", "warning", "error", "critical"]

        # add unique identifier of this psiturk project folder
        project_hash = hashlib.sha1(os.getcwd().encode()).hexdigest()[:12]
        self.user_options = {
            'bind': config.get('Server Parameters', "host") + ":" + config.get('Server Parameters', "port"),
            'workers': workers,
            'worker_class': 'gevent',
            'loglevels': self.loglevels,
            'loglevel': self.loglevels[config.getint("Server Parameters", "loglevel")],
            'accesslog': config.get('Server Parameters', "accesslog"),
            'errorlog': config.get('Server Parameters', "errorlog"),
            'proc_name': 'psiturk_experiment_server_' + project_hash,
            'limit_request_line': '0',
        }

        if config.get("Server Parameters", "certfile") and config.get("Server Parameters", "keyfile"):
            print("Loading SSL certs for server...")
            ssl_options = {
                'certfile': config.get('Server Parameters', "certfile"),
                'keyfile': config.get('Server Parameters', "keyfile")
            }
            self.user_options.update(ssl_options)

        if config.has_option("Server Parameters", "server_timeout"):
            self.user_options.update(
                {'timeout': config.get('Server Parameters', "server_timeout")})


def launch(app_dir=None):
    ExperimentServer(app_dir).run()


if __name__ == "__main__":
    launch()



from __future__ import generator_stop
import logging
logger = logging.getLogger('apscheduler')


class TaskUtils:
    _scheduler_aws_services_wrapper = None

    @property
    def aws_services_wrapper(self):
        if not self._scheduler_aws_services_wrapper:
            from .amt_services_wrapper import MTurkServicesWrapper
            self._scheduler_aws_services_wrapper = MTurkServicesWrapper()
        return self._scheduler_aws_services_wrapper


task_utils = TaskUtils()


def do_campaign_round(campaign, **kwargs):
    from .models import Participant
    from .experiment import app
    from .db import db_session

    # cancel if codeversion changes
    config_code_version = task_utils.aws_services_wrapper.config['Task Parameters']['experiment_code_version']
    if config_code_version != campaign.codeversion:
        logger.info(f'Codeversion changed (campaign: {campaign.codeversion}, config {config_code_version}), removing job.')
        return app.apscheduler.remove_job(kwargs['job_id'])

    # cancel if campaign goal met
    complete_count = Participant.count_completed(codeversion=campaign.codeversion, mode=campaign.mode)
    if complete_count >= campaign.goal:
        logger.info(f'Campaign goal met ({campaign.goal}), removing job.')
        campaign.end()
        db_session.add(campaign)
        db_session.commit()
        return

    task_utils.aws_services_wrapper.set_mode(campaign.mode)

    # how many for this round?
    all_hits = task_utils.aws_services_wrapper.get_all_hits().data
    available_count = task_utils.aws_services_wrapper.count_available(hits=all_hits).data
    pending_count = task_utils.aws_services_wrapper.count_pending(hits=all_hits).data

    maybe_will_complete_count = available_count + pending_count

    campaign_remaining = campaign.goal - maybe_will_complete_count - complete_count
    round_remaining = campaign.assignments_per_round
    remaining = min(campaign_remaining, round_remaining)
    logger.info(f'Posting total of {remaining} assignments this round.')
    while remaining:
        this_hit = min(remaining, 9)  # max 9 to avoid steep 40% commission
        result = task_utils.aws_services_wrapper.create_hit(
            num_workers=this_hit, reward=campaign.hit_reward, duration=campaign.hit_duration_hours)
        logger.info(result)
        remaining = remaining - this_hit
    return


# Approve All task_utils
def do_approve_all(mode):
    task_utils.aws_services_wrapper.set_mode(mode)
    result = task_utils.aws_services_wrapper.approve_all_assignments()
    if result.success:
        logger.info('Approved all.')
    else:
        logger.error(str(result.exception))



from __future__ import generator_stop
import os
import subprocess
import shutil


def _check_heroku_cmd_found():
    command = 'heroku'
    if not shutil.which(command):
        from psiturk.psiturk_exceptions import HerokuCmdNotFound
        raise HerokuCmdNotFound()


def _check_heroku_logged_in():
    if subprocess.getstatusoutput('heroku auth:whoami')[0]:
        from psiturk.psiturk_exceptions import HerokuNotLoggedIn
        raise HerokuNotLoggedIn()


def _check_is_git_repo():
    if subprocess.getstatusoutput('git rev-parse --git-dir')[0]:
        from psiturk.psiturk_exceptions import HerokuNotAGitRepo
        raise HerokuNotAGitRepo()


def _check_heroku_app_set():
    if subprocess.getstatusoutput('heroku config')[0]:
        from psiturk.psiturk_exceptions import HerokuAppNotSet
        raise HerokuAppNotSet()


def _set_heroku_config_vars():
    from psiturk.psiturk_config import PsiturkConfig

    CONFIG = PsiturkConfig()
    CONFIG.load_config()

    subprocess.call(['heroku', 'config:set', 'ON_CLOUD=true'])

    print()
    print('psiturk has finished setting heroku config vars.')


def _add_postgresql_db():
    subprocess.call(['heroku', 'addons:create', 'heroku-postgresql:hobby-dev'])
    print()
    print("\n".join([
        "Your psiturk heroku app will use the just-added postgresql database instead of whatever is set in your config.txt.",
        "If you don't want this, then change or unset the heroku config var 'DATABASE_URL'",
        "and/or remove the heroku postgresql addon."]))


HEROKU_FILES_DIR = os.path.join(os.path.dirname(__file__), "heroku_files")


def _copy_heroku_files():
    """
    Heroku needs a Procfile, a requirements.txt, and a file that stores the
    """
    _TARGET = os.curdir
    from distutils import dir_util
    print("Copying {} to {}".format(HEROKU_FILES_DIR, _TARGET))
    dir_util.copy_tree(HEROKU_FILES_DIR, _TARGET)
    print('Remember to commit these new files to your git repository.')


def do_heroku_setup():
    _check_heroku_cmd_found()
    _check_heroku_logged_in()
    _check_is_git_repo()
    _check_heroku_app_set()
    _set_heroku_config_vars()
    _copy_heroku_files()
    # _add_postgresql_db()

    print()
    print("Heroku config done.")
    print()
    print("Don't forget that your database needs to be a persistent datastore!")


if __name__ == '__main__':
    do_heroku_setup()
